/* XTV.c */
/*
========================================================================
Copyright (c) 2012 by Nigel Batten.

This work is licensed under the Creative Commons Attribution 3.0
Unported License. To view a copy of this license, visit
		http://creativecommons.org/licenses/by/3.0/
========================================================================
*/

#include "XTV.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "delay_x.h"

/*
	Monochrome composite TV signal generation.
	==========================================

	Overview:
	Two pins are used. One Sync pin, and one pixel pin.
	A resistor in series with each pin is used in conjuction with the
	75ohm resistor in the TV set to provide the following voltages:

		Sync:	0v
		Black:	0.3v
		White:	1.0v

	The Sync pattern is generated by a Scanline timer and a Compare
	/Match point. Modifying the Compare/Match point changes the width of
	the sync pulse at the start of the scanline.

	The Pixel pattern is generated by a USART peripheral in SPI mode.
	This generates a stream of pixels for a single scanline.
	The pixels are generated in a Compare/Match interrupt handler, which
	is configured to fire part-way through a display scanline.

	The pixels to display are found by looking up font data in flash.

	Optionally, DMA transfer can be used to pump data to the USART.
	This is of benefit, even though the pixel pattern is generated
	on-the-fly, because the pixel data can be generated faster than it
	needs to be transmitted. Handing the transmission pumping over to
	the DMA controller means that the CPU can go and do more useful
	things.

	Different sections of the signal require different SYNC signals.
	These sections are set-up by having a Scanline *counter* (clocked
	from the overflow of the Scanline *timer*) which fires a
	Compare/Match interrupt when the next section of the signal needs to
	be generated.

	The 'Render' interrupt is only fired during 'Display' portions of
	the signal.

	The actual peripherals used are defined in XTV.h, but their use
	is as follows:

	Timer/Counter usage
	===================

	2 timer/counters are used...
	SCANLINE_TIMER		this TIMER times each scanline
	SCANLINE_COUNTER	this COUNTER counts scanlines in the field/frame

	SCANLINE_TIMER	clocked 1:1 from system clock (e.g. 32Mhz)
		PER : Scanline Width (measured in sys clock cycles)
		CC<sync>   : Sync Width (pulse outputs to pin)
		CC<render> : DisplayStart (used for interrupt)

		DisplayStart interrupts are only enabled on 'display' scanlines

	SCANLINE_COUNTER	clocked from SCANLINE_TIMER overflow (via EventChannel 0)
		CC<pattern> : Next sync area start (Interrupt enabled)

	These Timer/Counters are configured in _XTV_InitSync(), below.

	The format of a signal is defined by a flash-based array of data,
	consisting of a number of SyncPattern entries, each of which
	described a section of the signal (scanline width, pulse width, 
	number of repetitions, whether it's a display area etc.)
	If a pattern entry is marked as 'display area', then the display
	interrupt will be enabled and subsequently fire part-way through the
	scanline.
*/

/* PUBLIC */
uint8_t	g_fieldcount ;


/* PRIVATE */
// Note: the first pattern in a format must be marked "FieldStart"
// Note: the last pattern in a format must be marked "LastPattern"
typedef enum SyncPatternInfo_enum {
	SyncPatternInfo_Normal_bm		= 0b00000000,
	SyncPatternInfo_Display_bm		= 0b00000001,
	SyncPatternInfo_FieldStart_bm	= 0b00000010,
	SyncPatternInfo_LastPattern_bm	= 0b10000000,
} SyncPatternInfo ;

typedef struct SyncPattern_struct{
	uint16_t	RepeatCount ;	// number of scanlines that this sync pattern should run for
	uint16_t	SyncWidth ;		// number of timer ticks in the sync pulse
	uint16_t	ScanlineWidth ;	// number of timer ticks in the scanline width
	SyncPatternInfo	Info ;
} SyncPattern_t ;

typedef struct SignalFormat_struct {
	char			Name[4] ;
	SyncPattern_t	Pattern[] ;
} SignalFormat_t ;

extern const SignalFormat_t f_SignalFormat_PAL		PROGMEM ; // PAL output.  Fake progressive. 288p50
extern const SignalFormat_t f_SignalFormat_NTSC		PROGMEM ; // NTSC output. Fake progressive. 240p60
extern const SignalFormat_t f_SignalFormat_576i		PROGMEM ; // PAL output.  Interlaced. 576i50
extern const SignalFormat_t f_SignalFormat_NTSCi	PROGMEM ; // NTSC output. Interlaced. 480i60

// current signal format and index into the sync pattern.
static volatile const SignalFormat_t* m_pSignalFormat = &f_SignalFormat_PAL ;
static volatile uint8_t m_SyncPatternIndex = 0 ;

// Macro to convert number of microseconds to a number of ticks on a timer.
#define US_TO_TICKS( TIME_US ) (((F_CPU / 1000000L) * TIME_US ) - 1)
const SignalFormat_t f_SignalFormat_PAL PROGMEM =
{
#define SCANLINE		US_TO_TICKS( 64 )
#define HSYNC			US_TO_TICKS( 4.37 )
#define HALFSCANLINE	(SCANLINE / 2)
#define SHORTSYNC		(HSYNC / 2)
#define BROADSYNC		(HALFSCANLINE - HSYNC)
	{'P','A','L',' '},
	{
		{ 47,	HSYNC,		SCANLINE,		SyncPatternInfo_FieldStart_bm },
		{ 258,	HSYNC,		SCANLINE,		SyncPatternInfo_Display_bm },
		{ 6,	SHORTSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },	// pre-equalisation pulses
		{ 5,	BROADSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },	// VSYNC proper
		{ 5,	SHORTSYNC,	HALFSCANLINE,	SyncPatternInfo_LastPattern_bm }// post-equalisation pulses
	}
#undef SCANLINE
#undef HALFSCANLINE
#undef HSYNC
#undef SHORTSYNC
#undef BROADSYNC
} ;

const SignalFormat_t f_SignalFormat_576i PROGMEM =
{
#define SCANLINE		US_TO_TICKS( 64 )
#define HSYNC			US_TO_TICKS( 4.37 )
#define HALFSCANLINE	(SCANLINE / 2)
#define SHORTSYNC		(HSYNC / 2)
#define BROADSYNC		(HALFSCANLINE - HSYNC)
	{'P','A','L','i'},
	{
		{ 48,	HSYNC,		SCANLINE,		SyncPatternInfo_FieldStart_bm },
		{ 257,	HSYNC,		SCANLINE,		SyncPatternInfo_Display_bm }, // starts on linecount 28
		{ 5,	SHORTSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },	// pre-equalisation pulses
		{ 5,	BROADSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },	// VSYNC proper
		{ 4,	SHORTSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },	// post-equalisation pulses
		{ 1,	SHORTSYNC,	SCANLINE,		SyncPatternInfo_FieldStart_bm },	// post-equalisation pulses
		{ 47,	HSYNC,		SCANLINE,		SyncPatternInfo_Normal_bm },
		{ 257,	HSYNC,		SCANLINE,		SyncPatternInfo_Display_bm }, // starts on 347
		{ 1,	HSYNC,		HALFSCANLINE,	SyncPatternInfo_Normal_bm },
		{ 5,	SHORTSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },
		{ 5,	BROADSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },// VSYNC proper
		{ 5,	SHORTSYNC,	HALFSCANLINE,	SyncPatternInfo_LastPattern_bm },// post-equalisation pulses
	}
#undef SCANLINE
#undef HALFSCANLINE
#undef HSYNC
#undef SHORTSYNC
#undef BROADSYNC
} ;

const SignalFormat_t f_SignalFormat_NTSC PROGMEM =
{
#define SCANLINE		US_TO_TICKS( 63.5 )
#define HSYNC			US_TO_TICKS( 4.37 )
#define HALFSCANLINE	(SCANLINE / 2)
#define SHORTSYNC		(HSYNC / 2)
#define BROADSYNC		(HALFSCANLINE - HSYNC)
	{'N','T','S','C'},
	{
		{ 20,	HSYNC,		SCANLINE,		SyncPatternInfo_FieldStart_bm },// 
		{ 234,	HSYNC,		SCANLINE,		SyncPatternInfo_Display_bm },	// Display area
		{ 6,	SHORTSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },	// pre-equalisation pulses
		{ 6,	BROADSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },	// VSYNC proper
		{ 6,	SHORTSYNC,	HALFSCANLINE,	SyncPatternInfo_LastPattern_bm }// post-equalisation pulses
	}
#undef SCANLINE
#undef HALFSCANLINE
#undef HSYNC
#undef SHORTSYNC
#undef BROADSYNC
} ;


const SignalFormat_t f_SignalFormat_NTSCi PROGMEM =
{
#define SCANLINE		US_TO_TICKS( 63.5 )
#define HSYNC			US_TO_TICKS( 4.37 )
#define HALFSCANLINE	(SCANLINE / 2)
#define SHORTSYNC		(HSYNC / 2)
#define BROADSYNC		(HALFSCANLINE - HSYNC)
	{'N','T','S','i'},
	{
		{ 20,	HSYNC,		SCANLINE,		SyncPatternInfo_FieldStart_bm },
		{ 233,	HSYNC,		SCANLINE,		SyncPatternInfo_Display_bm },	// Display area
		{ 1,	HSYNC,		HALFSCANLINE,	SyncPatternInfo_Normal_bm },
		{ 6,	SHORTSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },	// pre-equalisation pulses
		{ 6,	BROADSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },	// VSYNC proper
		{ 5,	SHORTSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },	// post-equalisation pulses
		{ 1,	SHORTSYNC,	SCANLINE,		SyncPatternInfo_Normal_bm },	// post-equalisation pulses
		{ 20,	HSYNC,		SCANLINE,		SyncPatternInfo_FieldStart_bm },
		{ 233,	HSYNC,		SCANLINE,		SyncPatternInfo_Display_bm },	// Display area
		{ 6,	SHORTSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },	// pre-equalisation pulses
		{ 6,	BROADSYNC,	HALFSCANLINE,	SyncPatternInfo_Normal_bm },	// VSYNC proper
		{ 6,	SHORTSYNC,	HALFSCANLINE,	SyncPatternInfo_LastPattern_bm }	// post-equalisation pulses
	}
#undef SCANLINE
#undef HALFSCANLINE
#undef HSYNC
#undef SHORTSYNC
#undef BROADSYNC
} ;

static const SignalFormat_t *f_SignalFormat[] PROGMEM = {
	&f_SignalFormat_PAL,
	&f_SignalFormat_NTSC,
	&f_SignalFormat_576i,
	&f_SignalFormat_NTSCi
} ;

/*
Build up the various definitions based on the configuration in XTV.h
*/

#define _CONCAT( A, B ) A ## B
#define _CONCAT2( A, B ) A ## B
#define _CONCAT3( A, B, C ) A ## B ## C
#define _CONCAT4( A, B, C, D ) A ## B ## C ## D
#define _CONCAT5( A, B, C, D, E ) A ## B ## C ## D ## E
#define _CONCAT6( A, B, C, D, E, F ) A ## B ## C ## D ## E ## F
#define CONCAT( A, B ) _CONCAT( A, B )
#define CONCAT2( A, B ) _CONCAT2(A, B )
#define CONCAT3( A, B, C ) _CONCAT3( A, B, C )
#define CONCAT4( A, B, C, D ) _CONCAT4( A, B, C, D )
#define CONCAT5( A, B, C, D, E ) _CONCAT5( A, B, C, D, E )
#define CONCAT6( A, B, C, D, E, F ) _CONCAT6( A, B, C, D, E, F )

// For some reason, these aren't defined in iox<blah>.h
#define USART_UDORD_bm   0x04	/* Data Order bit mask. */
#define USART_UDORD_bp   2		/* Data Order bit position */
#define USART_UCPHA_bm   0x02	/* Clock Phase bit mask. */
#define USART_UCPHA_bp   1		/* Clock Phase bit position. */

// sanity-check the configuration

// Convert the PORT and CompareChannel definitions to integers, so
// they can be compared by the preprocessor and hence flag problems
#define PORT_TO_INTEGER_A 1
#define PORT_TO_INTEGER_B 2
#define PORT_TO_INTEGER_C 3
#define PORT_TO_INTEGER_D 4
#define PORT_TO_INTEGER_E 5
#define PORT_TO_INTEGER_F 6
#define PORT_TO_INTEGER_G 7
#define PORT_TO_INTEGER_H 8
#define PORT_TO_INTEGER_J 9
#define PORT_TO_INTEGER_K 10
#define PORT_TO_INTEGER_Q 11
#define PORT_TO_INTEGER_R 12
#if 	( CONCAT2( PORT_TO_INTEGER_, SCANLINE_TIMER_PORT ) == CONCAT2( PORT_TO_INTEGER_, SCANLINE_COUNTER_PORT )) \
	&&	( SCANLINE_TIMER_INSTANCE == SCANLINE_COUNTER_INSTANCE )
#error The SCANLINE_TIMER and SCANLINE_COUNTER are configured to be the same Timer/Counter! They must be defined as different Timer/Counters.
#endif

#define TC_COMPARE_CHANNEL_TO_INTEGER_A 1
#define TC_COMPARE_CHANNEL_TO_INTEGER_B 2
#define TC_COMPARE_CHANNEL_TO_INTEGER_C 3
#define TC_COMPARE_CHANNEL_TO_INTEGER_D 4
#if ( CONCAT2( TC_COMPARE_CHANNEL_TO_INTEGER_, SCANLINE_TIMER_SYNC_COMPARE ) == CONCAT2( TC_COMPARE_CHANNEL_TO_INTEGER_, SCANLINE_TIMER_RENDER_COMPARE ) )
#error The SCANLINE_TIMERs SYNC_COMPARE channel and RENDER_COMPARE channel are the same. They must be defined as different channels.
#endif

// determine the pin used for a given (Timer/Counter,Compare Channel) tuple.
#define TC_INSTANCE_CHANNEL_PIN_0A 0
#define TC_INSTANCE_CHANNEL_PIN_0B 1
#define TC_INSTANCE_CHANNEL_PIN_0C 2
#define TC_INSTANCE_CHANNEL_PIN_0D 3
#define TC_INSTANCE_CHANNEL_PIN_1A 4
#define TC_INSTANCE_CHANNEL_PIN_1B 5
#define SCANLINE_TIMER_SYNC_PIN_bp CONCAT3( TC_INSTANCE_CHANNEL_PIN_, SCANLINE_TIMER_INSTANCE, SCANLINE_TIMER_SYNC_COMPARE )

// determine the TX pin used for a given USART instance.
#define USART_INSTANCE_TX_PIN_0 3
#define USART_INSTANCE_TX_PIN_1 7
#define PIXEL_TX_PIN_bp CONCAT2( USART_INSTANCE_TX_PIN_, PIXEL_USART_INSTANCE )

// Check that the sync pin and pixel pin aren't the same!
#if (CONCAT2(PORT_TO_INTEGER_,SCANLINE_TIMER_PORT) == CONCAT2(PORT_TO_INTEGER_,PIXEL_USART_PORT))\
			&& (PIXEL_TX_PIN_bp == SCANLINE_TIMER_SYNC_PIN_bp)
#error The TXD pin of the USART and the SYNC PIN are the same pin. A different configuration must be used.
#endif

// e.g. TCC0_CCB_vect
#define RENDER_INTERRUPT			CONCAT6( TC, SCANLINE_TIMER_PORT, SCANLINE_TIMER_INSTANCE, _CC, SCANLINE_TIMER_RENDER_COMPARE, _vect )

// e.g. TCC0
 #define SCANLINE_TIMER				CONCAT3( TC, SCANLINE_TIMER_PORT, SCANLINE_TIMER_INSTANCE )
// e.g. TCC1
 #define SCANLINE_COUNTER			CONCAT3( TC, SCANLINE_COUNTER_PORT , SCANLINE_COUNTER_INSTANCE )

// e.g. PORTC
#define SYNC_PORT					CONCAT2( PORT, SCANLINE_TIMER_PORT )
#define SYNC_PIN_bm					(1 << SCANLINE_TIMER_SYNC_PIN_bp)
// e.g. PIN0CTRL
#define SYNC_PINCTRL				CONCAT3( PIN, SCANLINE_TIMER_SYNC_PIN_bp, CTRL )
// e.g. CCA
#define SYNC_CC						CONCAT2( CC, SCANLINE_TIMER_SYNC_COMPARE )
// e.g. CCABUF
#define SYNC_CCBUF					CONCAT2( SYNC_CC, BUF )
// e.g. TC0_CAEN_bm
#define SYNC_ENABLE_bm				CONCAT5( TC, SCANLINE_TIMER_INSTANCE, _ , SYNC_CC, EN_bm )
// e.g. TC0_CCABV_bm  (compare channel A buffer valid bit mask)
#define SYNC_CCBV_bm				CONCAT5( TC, SCANLINE_TIMER_INSTANCE, _CC, SCANLINE_TIMER_SYNC_COMPARE, BV_bm )
// e.g. TC0_PERBV_bm
#define SYNC_PERBV_bm				CONCAT3( TC, SCANLINE_TIMER_INSTANCE, _PERBV_bm )

// e.g. CCB
#define RENDER_CC					CONCAT2( CC, SCANLINE_TIMER_RENDER_COMPARE )
// e.g. TC_CCBIF_bm
#define RENDER_CCIF_bm				CONCAT5( TC, SCANLINE_TIMER_INSTANCE, _, RENDER_CC, IF_bm )
// e.g. TC_CCBINTLVL_HI_gc
#define RENDER_CCINTLVL_gc			CONCAT3( TC_ , RENDER_CC, INTLVL_HI_gc )
// e.g. TC0_CCBINTLVL_gm
#define RENDER_CCINTLVL_gm			CONCAT5( TC, SCANLINE_TIMER_INSTANCE, _, RENDER_CC, INTLVL_gm )
// e.g. EVSYS_CHMUX_TCC0_OVF_gc
#define SCANLINE_TIMER_OVERFLOW		CONCAT5( EVSYS_CHMUX_ , TC, SCANLINE_TIMER_PORT, SCANLINE_TIMER_INSTANCE, _OVF_gc )
// e.g. CCA
#define PATTERN_CC					CONCAT2( CC, SCANLINE_COUNTER_PATTERN_COMPARE )


// e.g. TCC1_CCA_vect
#define PATTERN_CHANGE_INTERRUPT	CONCAT6( TC, SCANLINE_COUNTER_PORT , SCANLINE_COUNTER_INSTANCE, _ , PATTERN_CC, _vect )
// e.g. TC_CCAINTLVL_MED_gc
#define PATTERN_CHANGE_CCINTLVL_gc	CONCAT3( TC_, PATTERN_CC, INTLVL_MED_gc )
// e.g. TC1_CCAINTLVL_gm
#define PATTERN_CHANGE_CCINTLVL_gm	CONCAT5( TC, SCANLINE_COUNTER_INSTANCE, _, PATTERN_CC, INTLVL_gm )

// e.g. USARTC0
#define PIXEL_USART			CONCAT3( USART, PIXEL_USART_PORT, PIXEL_USART_INSTANCE )
// e.g. PORTC
#define PIXEL_PORT			CONCAT2( PORT, PIXEL_USART_PORT )
#define PIXEL_TX_PIN_bm		(1<<PIXEL_TX_PIN_bp)
// e.g. PIN3CTRL
#define PIXEL_TX_PIN_CTRL	CONCAT3( PIN, PIXEL_TX_PIN_bp, CTRL )

// e.g. TCC0_CCB_vect
#define RENDER_INTERRUPT CONCAT6( TC, SCANLINE_TIMER_PORT, SCANLINE_TIMER_INSTANCE, _CC, SCANLINE_TIMER_RENDER_COMPARE, _vect )

// e.g. CCB
#define RENDER_CC		CONCAT2( CC, SCANLINE_TIMER_RENDER_COMPARE )


#ifdef XTV_DMA_CHANNEL
	#ifndef DMA_CTRL
		#error No DMA controller found on this XMega, therefore DMA cannot be used to push pixel data. Comment out the XTV_DMA_CHANNEL definition in XTV.h to drop back to CPU driven USART filling.
	#endif
	#define USE_DMA
	// e.g. DMA_CH_TRIGSRC_USARTC0_DRE_gc
	#define DMA_TRIGGER_SOURCE CONCAT4( DMA_CH_TRIGSRC_USART,  PIXEL_USART_PORT, PIXEL_USART_INSTANCE, _DRE_gc )
#endif

#define PATTERN_CHANGE_INTERRUPT_ENABLE		SCANLINE_COUNTER.INTCTRLB |= PATTERN_CHANGE_CCINTLVL_gc
#define PATTERN_CHANGE_INTERRUPT_DISABLE	SCANLINE_COUNTER.INTCTRLB &= ~PATTERN_CHANGE_CCINTLVL_gm

#define FONT_CHAR_HEIGHT 9
#define FONT_CHAR_COUNT 256

void setSignalFormat( const SignalFormatType_t t, uint16_t renderpoint )
{
	// ensure that the pattern change interrupt doesn't fire... (I don't want it firing half-way through...)
	PATTERN_CHANGE_INTERRUPT_DISABLE ;

	// find the address of the signal definition...
	m_pSignalFormat = (SignalFormat_t *)pgm_read_word( &f_SignalFormat[t] ) ;

	// copy the *first* SyncPattern from flash to local memory...
	// This is easier than a bunch of pgm_read_word/bytes later on.
	SyncPattern_t pattern ;
	memcpy_P( &pattern, (void*)&(m_pSignalFormat->Pattern[0]), sizeof( SyncPattern_t )) ; 

	// update the scanline timings.
	SCANLINE_TIMER.PER = pattern.ScanlineWidth ;
	SCANLINE_TIMER.SYNC_CC = pattern.SyncWidth ;
	SCANLINE_TIMER.RENDER_CC = renderpoint ;

	// Organise for the next sync pattern to be triggered at the right time...
	SCANLINE_COUNTER.PER = UINT16_MAX ;
	SCANLINE_COUNTER.PATTERN_CC = pattern.RepeatCount - 1 ;
	m_SyncPatternIndex = 0 ;

	// and now enable the pattern change interrupt ...
	PATTERN_CHANGE_INTERRUPT_ENABLE ;
}

void _XTV_InitSync( const SignalFormatType_t t, uint16_t renderpoint ){
	// Configure the Scanline Timer...
	SCANLINE_TIMER.CTRLB |= SYNC_ENABLE_bm | TC_WGMODE_SS_gc ; // enable Sync compare channel output (to pin) and use single-slope PWM.
	SCANLINE_TIMER.CTRLA = TC_CLKSEL_DIV1_gc ; // start the timer, using SysClock / 1

	// Configure the Scanline Counter...
	EVSYS.CH0MUX = SCANLINE_TIMER_OVERFLOW ; // Connect Event Channel 0 to ScanlineTimer overflow
	SCANLINE_COUNTER.CTRLB |= TC_WGMODE_SS_gc ;		// use single-slope PWM
	SCANLINE_COUNTER.CTRLA = TC_CLKSEL_EVCH0_gc ;	// start the timer, using Event Channel 0.

	// Finally, configure the Sync output pin.
	SYNC_PORT.DIR |= SYNC_PIN_bm ;// set the sync pin to be an output
	SYNC_PORT.SYNC_PINCTRL |= PORT_INVEN_bm ; 		// invert the SYNC pin's logic.
	SYNC_PORT.SYNC_PINCTRL |= PORT_OPC_TOTEM_gc ;	// Totem pole. 1 = Vcc. 0 = Gnd

	// configure the Timer/Counter modules to handle the output signal's format...
	setSignalFormat( t, renderpoint ) ;
	// NB : configuring the pixel output is handled in init_Render()
	// NB : The DisplayRender interrupt routine is defined in Render.c
}


// This is called whenever the next sync pattern is needed in the field/frame.
ISR( PATTERN_CHANGE_INTERRUPT )
{
	// Make a local copy of the next sync pattern...
	SyncPattern_t pattern ;
	memcpy_P( &pattern, (void*)&(m_pSignalFormat->Pattern[ m_SyncPatternIndex ]), sizeof( SyncPattern_t )); 

	// now put the new sync width and scanline width into the SCANLINE_TIMER
	SCANLINE_TIMER.SYNC_CCBUF = pattern.SyncWidth ;
	SCANLINE_TIMER.PERBUF = pattern.ScanlineWidth ;
	SCANLINE_TIMER.CTRLGSET = SYNC_PERBV_bm | SYNC_CCBV_bm ; // they'll be updated at the end of the current scanline.

	uint8_t Info = pattern.Info ;
	// turn on or off the DisplayRender interrupt...
	if ( Info & SyncPatternInfo_Display_bm ) { // this is a rendering area of the field/frame
		// enable the display render interrupt
		SCANLINE_TIMER.INTFLAGS = RENDER_CCIF_bm ; // clear the interrupt flag so it doesn't fire *NOW*
		SCANLINE_TIMER.INTCTRLB |= RENDER_CCINTLVL_gc ;
	}
	else { // this area of the field/frame has no display
		// disable the display render interrupt
		SCANLINE_TIMER.INTCTRLB &= ~RENDER_CCINTLVL_gm ;
	}

	uint16_t repeatcount = pattern.RepeatCount ;
	if (Info & SyncPatternInfo_FieldStart_bm){	// this is the first sync pattern of the field/frame
		SCANLINE_COUNTER.CNT = 0 ;
		SCANLINE_COUNTER.PATTERN_CC = repeatcount - 1 ;
		g_fieldcount++ ;
	} else {	// this is a follow-on sync pattern.
		SCANLINE_COUNTER.PATTERN_CC += repeatcount ;
	}

	if ( Info & SyncPatternInfo_LastPattern_bm ){ // this is the last sync pattern
		m_SyncPatternIndex = 0 ;
	} else {
		m_SyncPatternIndex++ ;
	}
}

DisplayData_t g_DisplayData ;

uint8_t f_font[] PROGMEM =
{
	// Yes, it's really, really odd to include a .c file. I know.
	// I'll find some birch twigs.
	#include "fontbank0.c"
};


#ifndef USE_DMA
static void render_char_4cpp( void ) ; // 4 CPU clocks per pixel. Write direct to USART.
#endif

#ifdef USE_DMA
static void render_dma_4cpp( void ) ; // 4 CPU clocks per pixel. DMA fills USART via buffer
#endif

void inline _XTV_InitDisplayData( void ) ;

static uint8_t *s_fontslice ;
static uint8_t s_slice ;
static uint8_t *s_chardata ;
static uint8_t s_row ;


static inline void _XTV_InitField( void ){
	s_fontslice = &f_font[0] ;
	s_slice = 0 ;
	s_chardata = &g_DisplayData.CharData.Cell[0][0] ;
	s_row = 0 ;
}

#ifdef USE_DMA
static uint8_t slice[ COL_COUNT_MAX ] ; // Data for scanline output

void _XTV_InitDMA( void ){
	// Configure the DMA controller
	DMA_CTRL |= DMA_ENABLE_bm | DMA_DBUFMODE_DISABLED_gc | DMA_PRIMODE_RR0123_gc ;

	// Bust length of one byte, single-shot.
	DMA.XTV_DMA_CHANNEL.CTRLA |= DMA_CH_BURSTLEN_1BYTE_gc | DMA_CH_SINGLE_bm ;
//	DMA_CH0_CTRLA |= DMA_CH_BURSTLEN_1BYTE_gc | DMA_CH_SINGLE_bm ;
	// Addressing...
	//		Source : Increment, and reset to original after transaction is completed.
	//		Target : Fixed, don't reload.
	DMA.XTV_DMA_CHANNEL.ADDRCTRL = DMA_CH_SRCRELOAD_TRANSACTION_gc | DMA_CH_SRCDIR_INC_gc | DMA_CH_DESTRELOAD_NONE_gc | DMA_CH_DESTDIR_FIXED_gc ;
	//	DMA_CH0_ADDRCTRL = DMA_CH_SRCRELOAD_NONE_gc | DMA_CH_SRCDIR_INC_gc | DMA_CH_DESTRELOAD_NONE_gc | DMA_CH_DESTDIR_FIXED_gc ;
	// trigger the next byte when the USART's Data Register is Empty
	DMA.XTV_DMA_CHANNEL.TRIGSRC = DMA_TRIGGER_SOURCE ;
	// One byte per block, repeating for 40 blocks. (each trigger completes one block, hence 40 triggers required)
	DMA.XTV_DMA_CHANNEL.TRFCNT = 1 ;

	// Source address
	DMA.XTV_DMA_CHANNEL.SRCADDR0=(((uint16_t) &slice[0]) >> 0 ) & 0xFF;
	DMA.XTV_DMA_CHANNEL.SRCADDR1=(((uint16_t) &slice[0]) >> 8 ) & 0xFF;
	DMA.XTV_DMA_CHANNEL.SRCADDR2=0;

	// Target address
	DMA.XTV_DMA_CHANNEL.DESTADDR0=(((uint16_t) &PIXEL_USART.DATA) >> 0 ) & 0xFF;
	DMA.XTV_DMA_CHANNEL.DESTADDR1=(((uint16_t) &PIXEL_USART.DATA) >> 8 ) & 0xFF;
	DMA.XTV_DMA_CHANNEL.DESTADDR2=0;
}

void render_dma_4cpp(){
uint8_t *chardata = s_chardata ;
uint8_t *fontslice = s_fontslice ;
uint8_t *buf = slice ;

	#define FILL_BUFFER_1_BYTE *buf++ = ~pgm_read_byte( (uint16_t)fontslice + *chardata++ )

	// populate the first two bytes of the output array...
	FILL_BUFFER_1_BYTE ;
	FILL_BUFFER_1_BYTE ;

	// now enable the DMA transfer... (we've got 2 bytes ready to go)
	DMA.XTV_DMA_CHANNEL.REPCNT = COL_COUNT_MAX ;
	DMA.XTV_DMA_CHANNEL.CTRLA = DMA_CH_ENABLE_bm | DMA_CH_REPEAT_bm | DMA_CH_BURSTLEN_1BYTE_gc | DMA_CH_SINGLE_bm ;

	// now fill in the remainder of the buffer.
	for ( uint8_t i = COL_COUNT_MAX - 2 ; i ; i-- ){
		// Possible speedups:
		// 1. Unroll the loop.
		// 2. Make f_font  __attribute__ ((aligned(256)))
		FILL_BUFFER_1_BYTE ;
	}

	// setup the fontslice pointer for the next scanline...
	s_fontslice += 256;
	s_slice++ ;
	if (s_slice > 8){
		s_row++ ;
		if (s_row >= ROW_COUNT_MAX){
			// no more rows to draw in this field
			// disable the render interrupt.
			SCANLINE_TIMER.INTCTRLB &= ~RENDER_CCINTLVL_gm ;
			// initialise the data ready for the next field.
			_XTV_InitField() ;
		} else {
			// more rows to draw
			s_chardata = chardata ; // chardata will be pointing at the right place
			s_slice = 0 ;
			s_fontslice = &f_font[0] ; // first slice of font data.
		}
	}
}
#endif

#ifndef USE_DMA
void render_char_4cpp( void )
{
uint8_t *chardata = s_chardata ;
uint8_t *fontslice = s_fontslice ;

	for( uint8_t i = COL_COUNT_MAX ; i ; i-- )
	{
		uint8_t c = *chardata++ ;
		uint8_t *offset = fontslice + c ;
		uint8_t pixels1 = ~pgm_read_byte( offset ) ;
		PIXEL_USART.DATA = pixels1 ;
		_delay_cycles(18);
	}
	// setup the fontslice pointer for the next scanline...
	s_fontslice += 256;
	s_slice++ ;
	if (s_slice > 8){
		s_row++ ;
		if (s_row >= ROW_COUNT_MAX){
			// no more rows to draw in this field
			// disable the render interrupt.
			SCANLINE_TIMER.INTCTRLB &= ~RENDER_CCINTLVL_gm ;
			// initialise the data ready for the next field.
			_XTV_InitField() ;
		} else {
			// more rows to draw
			s_chardata = chardata ; // chardata will be pointing at the right place
			s_slice = 0 ;
			s_fontslice = &f_font[0] ;
		}
	}
}
#endif

ISR( RENDER_INTERRUPT )
{
#define TWEAK 8
	// First remove any jitter!

	// This interrupt will have up to 4 cycles of jitter, depending on what instruction was
	// executing at the point of firing. This jitter looks awful on video signals.

	// Fortunately, we can adjust for it, because our scanline timer runs 1:1 with the CPU clock

	// TODO: Find out if larger XMegas have longer instructions. If so, this adjustment
	// routine will need to handle more than 4 cycles of jitter.

	uint8_t now = SCANLINE_TIMER.CNT & 0xff ;
	
	uint8_t dt = ((SCANLINE_TIMER.RENDER_CC & 0xff) - TWEAK) ;
	now += dt & 0b11 ;
	
    // The following assembler will introduce a delay of 4c to 7c, depending on the bottom two bits of 'now'.
	asm("\n\t"
		"sbrs	%0 ,1	; Skip next instruction if bit 1 is set. (1c if noskip, 2c if skipping)		\n\t"
		"lpm			; equiv. to 3c NOP			    											\n\t"
		"sbrs 	%0 ,0	; skip next instruction if bit 0 is set. (1c if noskip, 2c if skipping)		\n\t"
		"rjmp	.+0		; equiv. to 2c NOP															\n\t"
		:
		: "r" (now)
	) ;

	// now that we've adjusted for the inevitable interrupt call-time jitter, call the render routine.
#ifdef USE_DMA
	render_dma_4cpp();
#else
	render_char_4cpp() ;
#endif
//	(*g_scanlineRender)() ;
}

#define DEFAULT_HORZ_POS_TICK (uint16_t)(DEFAULT_HORZ_POS_US * 32L)

#define _QUOTE(A) #A
#define QUOTE(A) _QUOTE(A)
uint8_t debugdata1[] PROGMEM = QUOTE(PIXEL_TX_PIN_bp) ;
uint8_t debugdata2[] PROGMEM = QUOTE(PIXEL_TX_PIN_CTRL) ;
uint8_t debugdata3[] PROGMEM = QUOTE(PIXEL_PORT) ;
uint8_t debugdata4[] PROGMEM = QUOTE(PIXEL_TX_PIN_bm) ;

void inline XTV_init( const SignalFormatType_t sig, uint16_t renderpoint_uS )
{
	// Initialisation of the ScanlineTimer and ScanlineCounter are handled in init_DisplaySync.
	_XTV_InitSync( sig, renderpoint_uS ) ;
	_XTV_InitDisplayData() ;

	// configure the pixel output pin to be inverted and wired-or.
	// The inversion is so the idle-high SPI will idle low.
	// (and saves having to enable/disable SPI on every scanline)
	// The wired-or means resistor calculations are a lot easier.
	// (essentially pins with '0' output are high-z and can be ignored)
	PIXEL_PORT.DIRSET = PIXEL_TX_PIN_bm ; // set TXC0 to be output
	PIXEL_PORT.PIXEL_TX_PIN_CTRL |= PORT_INVEN_bm | PORT_OPC_WIREDOR_gc ; // inverted + drive to Vcc only
	PIXEL_PORT.OUTSET = PIXEL_TX_PIN_bm ; // Ensure that the pixel line is low.

	//Configure USARTC0 as Master SPI mode with 4 clocks per bit
	PIXEL_USART.BAUDCTRLA = 1 ;
	PIXEL_USART.CTRLC = USART_CMODE_MSPI_gc | USART_UCPHA_bm ;
	PIXEL_USART.CTRLB = USART_TXEN_bm ;
	

	
	_XTV_InitField() ;
#ifdef USE_DMA
	_XTV_InitDMA() ;
#endif
}

void inline _XTV_InitDisplayData( void ){
	// Fill the array with spaces.
	for( uint8_t row = 0 ; row < ROW_COUNT_MAX ; row++ ){
		for( uint8_t col = 0 ; col < COL_COUNT_MAX ; col++ ){
			g_DisplayData.CharData.Cell[ row ][ col ] = ' ' ;
		}
	}
}
