/* XTV.h */
/*
========================================================================
Copyright (c) 2012 by Nigel Batten.

This work is licensed under the Creative Commons Attribution 3.0
Unported License. To view a copy of this license, visit
		http://creativecommons.org/licenses/by/3.0/
========================================================================
*/
#ifndef __XTV_h
#define __XTV_h

#include <avr/io.h>

/*
Summary:
========
	Text-based TV Output demonstration for the XMega microcontrollers.

	40 x 25 black-and-white character text via a composite video signal.

	The font has Code-page 437 characters.
	The font data is as per the TellyMate fontbank data.
		See Shields -> fontbanks on the Batsocks download page for
		different fonts that could be used.

		http://www.batsocks.co.uk/downloads/index_downloads.htm

	All the processing is done in background, leaving the main loop free
	for other tasks. Simply update the array of characters as and when
	the screen needs to be changed.

	CPU usage is between 30% (for DMA version) and 60% (without DMA).

Hardware requirements:
======================
	XMega processor
	2 resistors.
	(optional but recommended) 16Mhz crystal.

Peripherals used:
=================
	2 Timer/Counters (any port, any instance)
	1 USART (any port, any instance)
	1 Event Channel
	(optional) 1 DMA channel

Configuration:
==============
	By default, the following peripherals are used:

		TC C0, using OC0A (PC0) for Sync output.
		TC C1
		USART C0, using TXD0 (PC3) for Pixel output.

	This should be usable on all XMegas.
	For details on altering this configuration, see below.

Usage:
======
	Replace the dreadful demo code inside main() with your own code!

	Characters that you place in the g_DisplayData.CharData.Cell[25][40]
	array will appear on the screen!

	The call to XTV_init() includes a parameter which determines the
	TV signal format to generate. This can be one of:
		SignalFormatType_PAL	=> PAL fake progressive (288p50)
		SignalFormatType_NTSC	=> NTSC fake progressive (240p60)
		SignalFormatType_PALi	=> PAL interlaced (576i50)
		SignalFormatType_NTSCi	=> NTSC interlaced (480i60)
	The second parameter determines the display's horizontal position.

*/

/*
PERIPHERAL CONFIGURATION
========================

SCANLINE_TIMER uses two compare channels.
	One for SYNC pulse output (uses the output pin)
	One for Render position interrupt.

SCANLINE_COUNTER uses one compare channel
	One to trigger the next signal pattern area

PIXEL_USART
	Uses TXD output pin.

XTV_DMA_CHANNEL (*optional*)
	Used to throw bytes at the USART to reduce CPU load a little.

Default configuration:
	SCANLINE TIMER : PORTC, TC0 : CCA for SYNC, CCB for render pos.
	SCANLINE_COUNTER : PORT C, TC1 : CCA for pattern counting.
	PIXEL_USART : PORT C, USART0
	XTV_DMA_CHANNEL : NOT DEFINED.

Default pinout:
	PC0 : Sync output
	PC3 : Pixel output

 */

// SCANLINE_TIMER
#define SCANLINE_TIMER_PORT				C
#define SCANLINE_TIMER_INSTANCE			0

#define SCANLINE_TIMER_SYNC_COMPARE		A
#define SCANLINE_TIMER_RENDER_COMPARE	B

// SCANLINE_COUNTER
#define SCANLINE_COUNTER_PORT				C
#define SCANLINE_COUNTER_INSTANCE			1

#define SCANLINE_COUNTER_PATTERN_COMPARE	A

// PIXEL_USART
#define PIXEL_USART_PORT		C
#define PIXEL_USART_INSTANCE	0

// Uncomment the following line to use DMA to shovel data to the USART.
// (assuming you have a DMA controller!)
//#define XTV_DMA_CHANNEL CH0

/*
===============================
END OF PERIPHERAL CONFIGURATION
*/


typedef enum SignalFormatType_enum {
	SignalFormatType_PAL,
	SignalFormatType_NTSC,
	SignalFormatType_PALi,
	SignalFormatType_NTSCi
} SignalFormatType_t ;

extern uint8_t	g_fieldcount ;
extern void XTV_init( const SignalFormatType_t t, uint16_t renderpoint_tick ) ;
extern void XTV_SetSignalFormat( const SignalFormatType_t  t, uint16_t renderpoint_tick ) ;

#define ROW_COUNT_MAX 25
#define COL_COUNT_MAX 40

typedef struct CharacterData_struct{
	uint8_t Cell[ ROW_COUNT_MAX ][ COL_COUNT_MAX ] ;
} CharacterData_t ;

typedef union DisplayData_struct {
	CharacterData_t	CharData;
} DisplayData_t ;

extern DisplayData_t g_DisplayData ;


#endif
