/*
========================================================================
Copyright (c) 2012 by Nigel Batten.

This work is licensed under the Creative Commons Attribution 3.0
Unported License. To view a copy of this license, visit
		http://creativecommons.org/licenses/by/3.0/
========================================================================
Demonstration of TV output from an Atmel XMega.

*	40x25 character array.
*	NTSC or PAL output.
*	Interlaced or (fake) progressive output.
*	Entirely interrupt driven.

The pins shown below are for the default configuration.
Different Timer/Counters and USART peripherals can be used.
See XTV.h for more details on configuration.

Required hardware
	Two resistors, RSync and RPixel.
	(Pins show default configuration)
        
    C0 ---[RSync ]----.
                      |
    C3 ---[RPixel]----+---- Composite signal

    Running with a 3.3v Vcc,

    RSync  = 680 or 750 ohms
    RPixel = 150 or 160 ohms

Optional hardware:
	16Mhz crystal + caps to stop jiggly output

Peripherals used:
	TC C0 : Scanline timer
	TC C1 : Scanline counter
	USART C0 : Pixel Stream
	EventChannel0 : Ticks scanline counter from scanline timer overflow.

Pins used:
	Port C, Pin 0 : Composite Sync output
	Port C, Pin 3 : Pixel output

The sync pin drives around 4mA.
The pixel pin drives around 12mA on whites.

The sync generation and pixel squirting takes around 60% of the CPU.
This can be reduced to around 30% by using DMA to charge the USART.
(See XTV.h for more details - it's very simple)

The rendering routine uses the highest interrupt priority level.
If other high-level interrupts are introduced, the display will not be
stable. Medium and Low priority level interrupts are fine.

Other:
Hans-Juergen Heinrichs' delay_x header file is used to provide
cycle-accurate delays.


*/
#include <avr/interrupt.h>
#include <util/delay.h>
#include "XTV.h"

#include <avr/pgmspace.h>

static inline void init( void );
static inline void init_SystemClock_Internal( void ) ;
static inline void init_SystemClock_External( void ) ;

int __attribute__((OS_main)) main( void )
{
	init();
	sei(); // enable interrupts globally
	PMIC.CTRL = PMIC_MEDLVLEN_bm | PMIC_HILVLEN_bm;


/* TODO: Replace this rubbish example with your own code! */
// This really just shows characters being put into the data array.
#define DELAY_US 5000
	uint8_t nextchar = 0 ;
	while(1){
		
		void drawbox( uint8_t r1, uint8_t r2, uint8_t c1, uint8_t c2 ){
			for( uint8_t c = c1 ; c <= c2 ; c++ ){
				g_DisplayData.CharData.Cell[ r1 ][ c ] = nextchar++ ;
				_delay_us( DELAY_US ) ;
			}
			for( uint8_t r = r1 + 1 ; r <= r2 ; r++ ){
				g_DisplayData.CharData.Cell[ r ][ c2 ] = nextchar++ ;
				_delay_us( DELAY_US ) ;
			}
			for( uint8_t c = c2 - 1 ; c > c1 ; c-- ){
				g_DisplayData.CharData.Cell[ r2 ][ c ] = nextchar++ ;
				_delay_us( DELAY_US ) ;
			}
			for( uint8_t r = r2 ; r > r1 ; r--){
				g_DisplayData.CharData.Cell[ r ][ c1 ] = nextchar++ ;
				_delay_us( DELAY_US ) ;
			}
		}

		while(1){
			for (uint8_t border = 0 ; border <= 12 ; border++ ){
				drawbox( border, 24 - border, border, 39 - border ) ;
			}
			for (uint8_t border = 11 ; border > 0 ; border-- ){
				drawbox( border, 24 - border, border, 39 - border ) ;
			}
		}
	}
/* END of rubbish example */
}



static inline void init( void )
{
// Switch to use a 32Mhz clock...
	init_SystemClock_Internal() ;
//	init_SystemClock_External() ; // Use this if you've got a 16Mhz crystal!

// Initialise the TV output
#define MICROSECONDS_TO_TICKS( TIME_US ) (((F_CPU / 1000000L) * TIME_US ) - 1)
//	XTV_init( SignalFormatType_PALi, MICROSECONDS_TO_TICKS(13.5) ) ;
//	XTV_init( SignalFormatType_PAL, MICROSECONDS_TO_TICKS(13.5) ) ;
	XTV_init( SignalFormatType_NTSCi, MICROSECONDS_TO_TICKS(12.2) ) ;
//	XTV_init( SignalFormatType_NTSC, MICROSECONDS_TO_TICKS(12.2) ) ;
}

static inline void init_SystemClock_External( void ){
// Use an external 16Mhz crystal and x 2 PLL to give a clock of 32Mhz

	// Enable the external oscillator
	OSC.XOSCCTRL = OSC_FRQRANGE_12TO16_gc | OSC_XOSCSEL_XTAL_16KCLK_gc ;
	OSC.CTRL |= OSC_XOSCEN_bm ;
	while( (OSC.STATUS & OSC_XOSCRDY_bm) == 0 ){} // wait until stable

	// Now configure the PLL to be eXternal OSCillator * 2
	OSC.PLLCTRL = OSC_PLLSRC_XOSC_gc | 2 ;
	OSC.CTRL |= OSC_PLLEN_bm ; // enable the PLL...
	while( (OSC.STATUS & OSC_PLLRDY_bm) == 0 ){} // wait until stable

	// And now switch to the PLL as the clocksource
	CCP = CCP_IOREG_gc; // protected write follows   
	CLK.CTRL = CLK_SCLKSEL_PLL_gc;
}

static inline void init_SystemClock_Internal( void ){
	// Use 2Mhz internal RC with 16x PLL
	// This appears to be the least jittery (around 15ns jitter) source.

	OSC.PLLCTRL = OSC_PLLSRC_RC2M_gc | 16 ;
	OSC.CTRL |= OSC_PLLEN_bm ; // enable the PLL...
	while( (OSC.STATUS & OSC_PLLRDY_bm) == 0 ){} // wait until it's stable

	// And now we can (finally) switch to the PLL as a clocksource.
	CCP = CCP_IOREG_gc;				// protected write follows   
	CLK.CTRL = CLK_SCLKSEL_PLL_gc;	// The System clock is now the PLL output (2MhzRC * 16)

}
